@@ -1298,6 +1298,39 @@ class dfa_drop( dfa_base, state_drop ):
     pass
 
 
+class dfa_post( dfa_base, state ):
+    """A cpppo.dfa that supports the collection of a FIFO of post-processing closures to invoke when
+    the current invocation of the DFA is complete.  This is sometimes required when one or more
+    states in the DFA requires use of the same DFA to process some input.
+
+    If <dfa_post>.lock.locked(), use <dfa_post>.post.append( <closure> ) to schedule the activity
+    for immediately after release of the <dfa_post>'s lock.
+
+    """
+    def __init__( self, *args, **kwds ):
+        self.post		= []
+        super( dfa_post, self ).__init__( *args, **kwds )
+
+    def __exit__( self, typ, val, tbk ):
+        """After release of the DFA's lock, execute all the post-processing callables."""
+        try:
+            return super( dfa_post, self ).__exit__( typ, val, tbk )
+        finally:
+            while True:
+                with self.lock:
+                    if not self.post:
+                        break
+                    closure	= self.post.pop( 0 )
+                try:
+                    log.info( "%s -- post-processing %s",
+                              self.name_centered(), misc.function_name( closure ))
+                    closure()
+                except Exception as exc:
+                    log.warning( "%s -- post-processing %s failed w/ exception %s\n%s",
+                                 self.name_centered(), misc.function_name( closure ),
+                                 repr( exc ), ''.join( traceback.format_exc() ))
+
+
 class regex( dfa ):
     """Takes a regex in string or greenery.lego/fsm form, and converts it to a
     dfa.  We need to specify what type of characters our greenery.fsm
