@@ -68,29 +68,29 @@ def test_logix_multiple():
     data.read_frag.elements	= 1000
     data.read_frag.offset	= 0
     
-    # Request maximum size limited
-    beg,end,endactual		= Obj.request_elements( Obj_a1, data, 'read_frag' )
+    # Reply maximum size limited
+    beg,end,endactual		= Obj.reply_elements( Obj_a1, data, 'read_frag' )
     assert beg == 0 and end == 125 and endactual == 1000 # DINT == 4 bytes
-    beg,end,endactual		= Obj.request_elements( Obj_a3, data, 'read_frag' )
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'read_frag' )
     assert beg == 0 and end == 250 and endactual == 1000 # INT == 2 bytes
 
     data.read_frag.offset	= 125*4 # OK, second request; begin after byte offset of first
-    beg,end,endactual		= Obj.request_elements( Obj_a1, data, 'read_frag' )
+    beg,end,endactual		= Obj.reply_elements( Obj_a1, data, 'read_frag' )
     assert beg == 125 and end == 250 and endactual == 1000 # DINT == 4 bytes
 
     # Request elements limited; 0 offset
     data.read_frag.elements	= 30
     data.read_frag.offset	= 0
-    beg,end,endactual		= Obj.request_elements( Obj_a3, data, 'read_frag' )
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'read_frag' )
     assert beg == 0 and end == 30 and endactual == 30 # INT == 2 bytes
 
     # Request elements limited; +'ve offset
     data.read_frag.elements	= 70
     data.read_frag.offset	= 80
-    beg,end,endactual		= Obj.request_elements( Obj_a3, data, 'read_frag' )
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'read_frag' )
     assert beg == 40 and end == 70 and endactual == 70 # INT == 2 bytes
 
-    # Request limited by size of data provided (Write Tag Fragmented)
+    # Request limited by size of data provided (Write Tag [Fragmented])
     data			= cpppo.dotdict()
     data.service		= Obj.WR_FRG_RPY
     data.path			= { 'segment': [ cpppo.dotdict( d )
@@ -101,7 +101,7 @@ def test_logix_multiple():
     data.write_frag.data	= [0] * 100 # 100 elements provided in this request
     data.write_frag.elements	= 200       # Total request is to write 200 elements
     data.write_frag.offset	= 16        # request starts 16 bytes in (8 INTs)
-    beg,end,endactual		= Obj.request_elements( Obj_a3, data, 'write_frag' )
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'write_frag' )
     assert beg == 8 and end == 108 and endactual == 200 # INT == 2 bytes
 
     # ... same, but lets say request started somewhere in the middle of the array
@@ -109,9 +109,83 @@ def test_logix_multiple():
                                                  for d in [
                                                          {'element': 222 },
                                                        ]] }
-    beg,end,endactual		= Obj.request_elements( Obj_a3, data, 'write_frag' )
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'write_frag' )
     assert beg == 8+222 and end == 108+222 and endactual == 200+222 # INT == 2 bytes
 
+    # Ensure correct computation of (beg,end] that are byte-offset and data/size limited
+    data			= cpppo.dotdict()
+    data.service		= Obj.WR_FRG_RPY
+    data.path			= { 'segment': [] }
+
+    data.write_frag		= {}
+    data.write_frag.data	= [3,4,5,6]
+    data.write_frag.offset	= 6
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'write_frag' )
+    assert beg == 3 and end == 7 and endactual == 1000 # INT == 2 bytes
+
+    # Trigger the error cases only accessible via write
+
+    # Too many elements provided for attribute capacity
+    data.write_frag.offset	= ( 1000 - 3 ) * 2
+    try:
+        beg,end,endactual	= Obj.reply_elements( Obj_a3, data, 'write_frag' )
+        assert False, "Should have raised Exception due to capacity"
+    except Exception as exc:
+        assert "capacity exceeded" in str( exc )
+
+    data			= cpppo.dotdict()
+    data.service		= Obj.RD_FRG_RPY
+    data.path			= { 'segment': [] }
+
+    data.read_frag		= {}
+    data.read_frag.offset	= 6
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'read_frag' )
+    assert beg == 3 and end == 253 and endactual == 1000 # INT == 2 bytes
+
+    # And we should be able to read with an offset right up to the last element
+    data.read_frag.offset	= 1998
+    beg,end,endactual		= Obj.reply_elements( Obj_a3, data, 'read_frag' )
+    assert beg == 999 and end == 1000 and endactual == 1000 # INT == 2 bytes
+
+
+    # Trigger all the remaining error cases
+
+    # Unknown service
+    data.service		= Obj.RD_FRG_REQ
+    try:
+        beg,end,endactual	= Obj.reply_elements( Obj_a3, data, 'read_frag' )
+        assert False, "Should have raised Exception due to service"
+    except Exception as exc:
+        assert "unknown service" in str( exc )
+
+    # Offset indivisible by element size
+    data.service		= Obj.RD_FRG_RPY
+    data.read_frag.offset	= 7
+    try:
+        beg,end,endactual	= Obj.reply_elements( Obj_a3, data, 'read_frag' )
+        assert False, "Should have raised Exception due to odd byte offset"
+    except Exception as exc:
+        assert "element boundary" in str( exc )
+
+    # Initial element outside bounds
+    data.read_frag.offset	= 2000
+    try:
+        beg,end,endactual	= Obj.reply_elements( Obj_a3, data, 'read_frag' )
+        assert False, "Should have raised Exception due to initial element"
+    except Exception as exc:
+        assert "initial element invalid" in str( exc )
+
+    # Ending element outside bounds
+    data.read_frag.offset	= 0
+    data.read_frag.elements	= 1001
+    try:
+        beg,end,endactual	= Obj.reply_elements( Obj_a3, data, 'read_frag' )
+        assert False, "Should have raised Exception due to ending element"
+    except Exception as exc:
+        assert "ending element invalid" in str( exc )
+
+    # Beginning element after ending (should be no way to trigger)
+
 
     # Test an example valid multiple request
     data			= cpppo.dotdict()
